'***********************************
'*  VGA 1024 x 768 - 16bit bitmap  *
'***********************************

' P4 = VSYNC
' P3 = RED
' P2 = GREEN
' P1 = BLUE
' P0 = HSYNC

CON

  _clkmode = xinput
  _xinfreq = 60_000_000

SDRAM_COMMANDS  =       $6000
SDRAM_BUFFER    =       $8000

  s = 44	'scales DAC output (s = 0..128)


DAT		org

		setctra	#%00111			'set ctra for PLL*8
                setfrqa	frqa_			'set frqa

		setvid	#%0100			'set vid for vga output
		setvidy	vidy_			'set yiq coefficients
		setvidi	vidi_
		setvidq	vidq_

		cfgdacs	#%11_11_11_11		'set dacs to video mode

		cfgpins	pinmask,pindacs		'set pins for dac outputs
		setquad	#quad0

                reps    #4,#1
                setptra sdram_par               'write initial command list
                wrlong  h00000000,ptra++        'skip_done then sdram_address (x2)
                wrlong  h00000008,ptra++        'terminate with end_of_list

                setcog  #8
                coginit sdram_adr,sdram_par     'launch SDRAM driver into cog1

'
'
' Field loop
'
field           mov	scrx,vFp			'top blanks
		call	#blank

                mov     scry,vLines		'set visible lines
		setspa	#0

		mov	sdram_ptr,#0
                wrlong  sdram_ptr,sdram_loc
                add     sdram_ptr,sdram_add
                wrlong  hubram_rd1,hubram_loc
                wrlong  sdram_ptr,sdram_loc2
                add     sdram_ptr,sdram_add
                wrlong  hubram_rd2,hubram_loc2

line
		call	#sync
		setptra	bitmapptr

		waitvid	mode,#0
'	        call	#get128
		reps	#256/8,#8
		nop
		rdquad	ptra++
		nop
		nop
		nop
		pusha	quad0
		pusha	quad1
		pusha	quad2
		pusha	quad3

		waitvid	mode,#$80
'	        call	#get128
		reps	#256/8,#8
		nop
		rdquad	ptra++
		nop
		nop
		nop
		pusha	quad0
		pusha	quad1
		pusha	quad2
		pusha	quad3

		waitvid	mode,#0
'	        call	#get128
		reps	#256/8,#8
		nop
		rdquad	ptra++
		nop
		nop
		nop
		pusha	quad0
		pusha	quad1
		pusha	quad2
		pusha	quad3

'		rdlong	tmp,hubram_loc
'		tjnz	tmp,#$-1
                wrlong  sdram_ptr,sdram_loc
                add     sdram_ptr,sdram_add
                wrlong  hubram_rd1,hubram_loc

		waitvid	mode,#$80
'	        call	#get128
		reps	#256/8,#8
		nop
		rdquad	ptra++
		nop
		nop
		nop
		pusha	quad0
		pusha	quad1
		pusha	quad2
		pusha	quad3

'		rdlong	tmp,hubram_loc2
'		tjnz	tmp,#$-1
                wrlong  sdram_ptr,sdram_loc2
                add     sdram_ptr,sdram_add
                wrlong  hubram_rd2,hubram_loc2

	        djnz    scry,#line           	'another line?

                mov     scrx,vBp			'bottom blanks
		call	#blank

		setp	#0			'do vertical sync
		mov	scrx,vSync
		call	#blank
		clrp	#0

                jmp     #field                  'field loop


blank		call	#sync			'blank lines
		waitvid	hPix,sync_color0
		djnz	scrx,#blank
blank_ret	ret

sync						'horizontal sync
sync_ret	retd
		waitvid	hFp,sync_color0
		waitvid	hSync,sync_color1
		waitvid	hBp,sync_color0

get128		reps	#248/8,#8
		nop
		rdquad	ptra++
		nop
		nop
		nop
		pusha	quad0
		pusha	quad1
		pusha	quad2
		pusha	quad3
		rdquad	ptra++
		nop
		nop
		nop
		pusha	quad0
get128_ret	retd
		pusha	quad1
		pusha	quad2
		pusha	quad3
'
'
' Initialized data
'
sdram_adr       long    @SDRAM_Driver + $E80
sdram_par       long    SDRAM_COMMANDS
h00000000       long    $00000000
h00000008       long    $00000008

hubram_loc	long    SDRAM_COMMANDS
sdram_loc	long    SDRAM_COMMANDS+4
hubram_loc2	long    SDRAM_COMMANDS+8
sdram_loc2	long    SDRAM_COMMANDS+12
hubram_rd1	long    SDRAM_BUFFER+7
hubram_rd2	long    SDRAM_BUFFER+$400+7
sdram_ptr	long    0
sdram_add	long    512*2

vidy_		long	$00_01_00_00 * s	'red gain
vidi_		long	$00_00_01_00 * s	'green gain
vidq_		long	$00_00_00_01 * s	'blue gain

pinmask		long	%1111			'pin mask for P0..P3
pindacs		long	%1111_101100000		'pin configuration for video DAC

sync_color0	long	$0_000000
sync_color1	long	$1_000000

scrx		long	0
scry		long	0

		'	color base	mode		clocks/pixel	clocks/waitvid
mode		long	$00 << 24 +	$0e << 20 +	1 << 13 +	256
frqa_		long	$22AA_0000

hFp		long	24
hSync		long	136
hBp		long	160
hPix		long	1024

vFp		long	3
vSync		long	6
vBp		long	29
vLines		long 	768

bitmapptr	long	SDRAM_BUFFER

quad0		res	1
quad1		res	1
quad2		res	1
quad3		res	1

tmp		res	1

'********************************
'*                              *
'*  Propeller II SDRAM Driver   *
'*     for 16M x 16 devices     *
'*        (FPGA version)        *
'*                              *
'*  Version 0.1                 *
'*  9 April 2013                *
'*  by Chip Gracey              *
'*                              *
'********************************

{
SDRAM connections:

        P85 = cke (held high)           Port C
        P84 = cs
        P83 = ras
        P82 = cas
        P81 = we
        P80 = udqm (not used)
        P79 = ldqm (not used)
   P78..P77 = ba[1..0]
   P76..P64 = a[12..0]

   P63..P48 = dq[15..0]                 Port B


  Note: All pin directions have a 2-clock delay
        All pin outputs going to SDRAM have a 3-clock delay, since they are registered at the pin
        All pin inputs coming from SDRAM have a 2-clock delay, since they are registered at the pin


Commands (pairs of longs):

name            quads   bytes   hub_address (+0, set 2nd)                       sdram_address (+1, set 1st)
------------------------------------------------------------------------------------------------------------------------
rw_1024         64      1024    %xxxx_xxxx_xxxx_xxxA_AAAA_AAAA_AAAA_W111        %xxxx_xxxA_AAAA_AAAA_AAAA_AA00_0000_0000
rw_512          32       512    %xxxx_xxxx_xxxx_xxxA_AAAA_AAAA_AAAA_W110        %xxxx_xxxA_AAAA_AAAA_AAAA_AAA0_0000_0000
rw_256          16       256    %xxxx_xxxx_xxxx_xxxA_AAAA_AAAA_AAAA_W101        %xxxx_xxxA_AAAA_AAAA_AAAA_AAAA_0000_0000
rw_128           8       128    %xxxx_xxxx_xxxx_xxxA_AAAA_AAAA_AAAA_W100        %xxxx_xxxA_AAAA_AAAA_AAAA_AAAA_A000_0000
rw_64            4        64    %xxxx_xxxx_xxxx_xxxA_AAAA_AAAA_AAAA_W011        %xxxx_xxxA_AAAA_AAAA_AAAA_AAAA_AA00_0000
rw_32            2        32    %xxxx_xxxx_xxxx_xxxA_AAAA_AAAA_AAAA_W010        %xxxx_xxxA_AAAA_AAAA_AAAA_AAAA_AAA0_0000
rw_16            1        16    %xxxx_xxxx_xxxx_xxxA_AAAA_AAAA_AAAA_W001        %xxxx_xxxA_AAAA_AAAA_AAAA_AAAA_AAAA_0000
skip_done        0         0    %0000_0000_0000_0000_0000_0000_0000_0000        %xxxx_xxxx_xxxx_xxxx_xxxx_xxxx_xxxx_xxxx
end_of_list      0         0    %0000_0000_0000_0000_0000_0000_0000_1000        %xxxx_xxxx_xxxx_xxxx_xxxx_xxxx_xxxx_xxxx


The driver scans a list for commands. Commands take two longs and are structured as hub_address + sdram_address. To have
the driver perform a read or write operation on the SDRAM, first set the sdram_address (2nd long) and then the hub_address
(1st long). The driver will set each hub_address long to skip_done (0) when its associated operation has completed.

Before starting the driver, build the command list structure with pairs of 0's, then terminate it with an 8. When launching
the driver into a cog, point the parameter (S) to the start of the command list.


Command list (longs):

        hub_address (skip_done/rw_xxxx)
        sdram_address

        hub_address (skip_done/rw_xxxx)
        sdram_address

        hub_address (skip_done/rw_xxxx)
        sdram_address

        ...

        end_of_list
}

DAT             org
'
'
' Initialize SDRAM                              clocks  hub     
'                                               ----------------
sdram_driver    getptra list                    '1      -               save command list address

                reps    #6000,#1                '1      -               repeat instruction for 100us @60MHz
                mov     pinc,h003FFFFF          '1      -               cke and cs high
                mov     dirc,h003FFFFF  wz      '6000   -               set SDRAM signals to outputs (100us), Z = 0
'
'
' Scan list for commands
'
:finish if_z    wrlong  cmd,ptra[-2]            '1      0               if command finished, set it to skip_done (cmd = 0)

                mov     pinc,h00240400          '1      1               issue 'precharge all' command
                setb    pinc,#20        wz      '1      2              deselect, Z = 0
:reset  if_z    setptra list                    '1      3               if end_of_list, reset list pointer; satisy Trp

                mov     pinc,h00200037          '1      4               issue 'set mode' command
                setb    pinc,#20                '1      5               deselect, satisfy Tmrd

                mov     pinc,h00226000          '1      6               issue 'auto refresh' command
                setb    pinc,#20                '1      7               deselect, satisfy Trfc in 8 more clocks

:next           rdlong  cmd,ptra++[2]   wz      '3      0..2            check command, point to next
                setptrb cmd                     '1      3               in case command, set hub address
        if_z    jmp     #:next                  '1/4    4/4..7          if skip_done, next command

                decod3  cmd             wc      '1      5               decode size bits, C = write
                and     cmd,#%11111110  wz      '1      6               isolate bits 7..1, Z = 1 if end_of_list
        if_z    jmp     #:reset                 '1/4    7/7..2          if end_of_list, reset list pointer
'
'
' Execute read/write command                    clocks  hub     XFR write SDRAM         XFR read SDRAM
'                                               ------------------------------------------------------
                rdlong  adr,ptra[-1]            '3      0..2    -                       -                       get SDRAM address

                shl     adr,#7                  '1      3       -                       -                       make 'active' command with bank and row address
                or      adr,#%1_0011_00         '1      4       -                       -                       (ba[1..0], a[12..0]) = adr[24..10]
                rol     adr,#15                 '1      5       -                       -

                mov     pinc,adr                '1      6       -                       -                       issue 'active' command
                setb    pinc,#20                '1      7       -                       -                       deselect, satisfy Trcd in 1 more clock

        if_c    rdquad  ptrb++                  '1      0       -                       -                       if write, read initial QUADs from hub

                setbc   :quad,#0                '1      1       -                       -                       set wrquad/rdquad according to read/write

        if_c    setxfr  #%010_011               '1      2       <QUADs_to_16_pins>      -                       if write, configure XFR at hub cycle 2

                shr     adr,#22+1               '1      3       -                       -                       make blank command with bank and column address
                and     pinc,h00306000          '1      4       -                       -                       (ba[1..0], a[12..0]) = (adr[24..23], %0000, adr[9..1])
                or      pinc,adr                '1      5       -                       -

        if_c    mov     dirb,hFFFF0000          '1      6       -                       -                       if write, enable data outputs

        if_nc   setxfr  #%100_011               '1      7                               <16_pins_to_QUADs>      if read, configure XFR at hub cycle 7

        if_nc   xor     pinc,h001A0000          '1      0       -                       -                       if read, issue 'read' command
        if_nc   setb    pinc,#20                '1      1       -                       -                       if read, deselect

        if_c    xor     pinc,h00180000          '1      2       -                       -                       if write, issue 'write' command (aligns with XFR on next clock)
        if_c    setb    pinc,#20                '1      3       output QUAD0 w0         -                       if write, deselect; read: SDRAM sees 'read' command

                shr     cmd,#1                  '1      4       output QUAD0 w1         -                       get loop count

        if_nc   nop     #7                      '1      5       output QUAD1 w0         -                       if read, pad time
                '(nop)                          '1      6       (skip to next hub 6)    -                       read: SDRAM starts outputting data stream
                '(nop)                          '1      7                               -
                '(nop)                          '1      0                               input w0                read: SDRAM data stream begins arriving in XFR
                '(nop)                          '1      1                               input w1 -> QUAD0
                '(nop)                          '1      2                               input w0
                '(nop)                          '1      3                               input w1 -> QUAD1
                '(nop)                          '1      4                               input w0

:quad           '(rdquad/wrquad)                '1      5       output QUAD1 w0         input w1 -> QUAD2
                '(rdquad/wrquad)                '1      6       output QUAD1 w1         input w0
                '(rdquad/wrquad)                '1      7       output QUAD2 w0         input w1 -> QUAD3
                rdquad  ptrb++                  '1      0       output QUAD2 w1         input w0                write: read next QUADS; read: write current QUADS
                djnz    cmd,#:quad              '1      1       output QUAD3 w0         input w1 -> QUAD0       loop for each set of QUADs
                '(djnz looping)                 '1      2       output QUAD3 w1         input w0
                '(djnz looping)                 '1      3       output QUAD0 w0 (new)   input w1 -> QUAD1       write: RDQUAD data valid
                '(djnz looping)                 '1      4       output QUAD0 w1         input w0

                mov     pinc,h002C0000          '1      2       output QUAD3 w1         -                       issue 'burst terminate' command (aligns with XFR on next clock)
                mov     dirb,#0         wz      '1      3       -                       -                       cancel data outputs (aligns with 'burst terminate'), Z=1
                jmp     #:finish                '4      4..7    -                       -                       finish up, scan for next command
'
' Constants
'
h00180000       long    $00180000               'write' command toggle
h001A0000       long    $001A0000               'read' command toggle
h00200037       long    $00200037               'set mode' command (full-page r/w bursts, cas latency = 3)
h00226000       long    $00226000               'read' command mask
h00240400       long    $00240400               'precharge all' command
h002C0000       long    $002C0000               'burst terminate' command
h00306000       long    $00306000               'blank command mask
h003FFFFF       long    $003FFFFF               'control pins mask
hFFFF0000       long    $FFFF0000               'data pins mask

'
' Variables
'
list            res     1                       'command list pointer
cmd             res     1                       'command
adr             res     1                       'address

'-------------------------------
